@@ -512,6 +512,15 @@ test_prompt_close (Test *test,
 	egg_assert_not_object (prompt);
 }
 
+static void
+on_prompt_close (GcrPrompt *prompt,
+                 gpointer user_data)
+{
+	gboolean *prompt_closed = (gboolean *)user_data;
+	g_assert (*prompt_closed == FALSE);
+	*prompt_closed = TRUE;
+}
+
 static void
 test_close_cancels (Test *test,
                     gconstpointer unused)
@@ -520,6 +529,7 @@ test_close_cancels (Test *test,
 	GError *error = NULL;
 	const gchar *password = NULL;
 	GAsyncResult *result = NULL;
+	gboolean prompt_closed;
 
 	gcr_mock_prompter_set_delay_msec (3000);
 	gcr_mock_prompter_expect_password_ok ("booo", NULL);
@@ -528,11 +538,15 @@ test_close_cancels (Test *test,
 	g_assert_no_error (error);
 	g_assert (GCR_IS_SYSTEM_PROMPT (prompt));
 
+	prompt_closed = FALSE;
+	g_signal_connect_after (prompt, "prompt-close", G_CALLBACK (on_prompt_close), &prompt_closed);
+
 	gcr_prompt_password_async (prompt, NULL, on_async_result, &result);
 
 	gcr_system_prompt_close (GCR_SYSTEM_PROMPT (prompt), NULL, &error);
 	g_assert_no_error (error);
 
+	g_assert (prompt_closed == TRUE);
 	egg_test_wait ();
 
 	password = gcr_prompt_password_finish (prompt, result, &error);
@@ -544,10 +558,178 @@ test_close_cancels (Test *test,
 	egg_assert_not_object (prompt);
 }
 
+static void
+test_close_from_prompter (Test *test,
+                          gconstpointer unused)
+{
+	GcrPrompt *prompt;
+	GError *error = NULL;
+	gboolean ret;
+	const gchar *password;
+	gboolean prompt_closed;
+
+	gcr_mock_prompter_expect_close ();
+
+	prompt = gcr_system_prompt_open_for_prompter (test->prompter_name, 1, NULL, &error);
+	g_assert_no_error (error);
+	g_assert (GCR_IS_SYSTEM_PROMPT (prompt));
+
+	prompt_closed = FALSE;
+	g_signal_connect_after (prompt, "prompt-close", G_CALLBACK (on_prompt_close), &prompt_closed);
+
+	ret = gcr_prompt_confirm_run (prompt, NULL, &error);
+	g_assert_no_error (error);
+	g_assert (ret == GCR_PROMPT_REPLY_CANCEL);
+
+	/* The prompt should be closed now, these shouldn't reach the mock prompter */
+
+	while (!prompt_closed)
+		g_main_context_iteration (NULL, TRUE);
+
+	ret = gcr_prompt_confirm_run (prompt, NULL, &error);
+	g_assert_no_error (error);
+	g_assert (ret == GCR_PROMPT_REPLY_CANCEL);
+
+	password = gcr_prompt_password_run (prompt, NULL, &error);
+	g_assert_no_error (error);
+	g_assert (password == NULL);
+
+	g_object_unref (prompt);
+	egg_assert_not_object (prompt);
+}
+
+static void
+test_after_close_dismisses (Test *test,
+                            gconstpointer unused)
+{
+	GcrPrompt *prompt;
+	GError *error = NULL;
+	gboolean ret;
+	const gchar *password;
+	gboolean prompt_closed;
+
+	gcr_mock_prompter_expect_confirm_ok (NULL);
+
+	prompt = gcr_system_prompt_open_for_prompter (test->prompter_name, 1, NULL, &error);
+	g_assert_no_error (error);
+	g_assert (GCR_IS_SYSTEM_PROMPT (prompt));
+
+	prompt_closed = FALSE;
+	g_signal_connect_after (prompt, "prompt-close", G_CALLBACK (on_prompt_close), &prompt_closed);
+
+
+	ret = gcr_prompt_confirm_run (prompt, NULL, &error);
+	g_assert_no_error (error);
+	g_assert (ret == GCR_PROMPT_REPLY_CONTINUE);
+
+	gcr_prompt_close (prompt);
+	g_assert (prompt_closed);
+
+	/* These should never even reach the mock prompter */
+
+	ret = gcr_prompt_confirm_run (prompt, NULL, &error);
+	g_assert_no_error (error);
+	g_assert (ret == GCR_PROMPT_REPLY_CANCEL);
+
+	password = gcr_prompt_password_run (prompt, NULL, &error);
+	g_assert_no_error (error);
+	g_assert (password == NULL);
+
+	while (g_main_context_iteration (NULL, FALSE));
+
+	g_object_unref (prompt);
+	egg_assert_not_object (prompt);
+}
+
+typedef struct {
+	GAsyncResult *result1;
+	GAsyncResult *result2;
+} ResultPair;
+
+static void
+on_result_pair_one (GObject *source,
+                    GAsyncResult *result,
+                    gpointer user_data)
+{
+	ResultPair *pair = user_data;
+	g_assert (pair->result1 == NULL);
+	pair->result1 = g_object_ref (result);
+	if (pair->result1 && pair->result2)
+		egg_test_wait_stop ();
+}
+
+static void
+on_result_pair_two (GObject *source,
+                    GAsyncResult *result,
+                    gpointer user_data)
+{
+	ResultPair *pair = user_data;
+	g_assert (pair->result2 == NULL);
+	pair->result2 = g_object_ref (result);
+	if (pair->result1 && pair->result2)
+		egg_test_wait_stop ();
+}
+
+static void
+test_watch_cancels (Test *test,
+                    gconstpointer unused)
+{
+	GDBusConnection *connection;
+	GcrPrompt *prompt;
+	GcrPrompt *prompt2;
+	GError *error = NULL;
+	const gchar *password;
+	ResultPair pair = { NULL, NULL };
+
+	gcr_mock_prompter_set_delay_msec (3000);
+	gcr_mock_prompter_expect_password_ok ("booo", NULL);
+
+	connection = g_bus_get_sync (G_BUS_TYPE_SYSTEM, NULL, &error);
+	g_assert_no_error (error);
+
+	/* This should happen immediately */
+	prompt = gcr_system_prompt_open_for_prompter (test->prompter_name, 0, NULL, &error);
+	g_assert_no_error (error);
+	g_assert (GCR_IS_SYSTEM_PROMPT (prompt));
+
+	/* Show a password prompt */
+	gcr_prompt_password_async (prompt, NULL, on_result_pair_one, &pair);
+
+	/* This prompt should wait, block */
+	gcr_system_prompt_open_for_prompter_async (test->prompter_name, 0, NULL,
+	                                           on_result_pair_two, &pair);
+
+	/* Wait a bit before stopping, so outgoing request is done */
+	g_usleep (G_TIME_SPAN_SECOND / 4);
+
+	/* Kill the mock prompter */
+	gcr_mock_prompter_disconnect ();
+
+	/* Both the above operations should cancel */
+	egg_test_wait ();
+
+	prompt2 = gcr_system_prompt_open_finish (pair.result2, &error);
+	g_assert_error (error, G_IO_ERROR, G_IO_ERROR_CANCELLED);
+	g_clear_error (&error);
+	g_assert (prompt2 == NULL);
+
+	password = gcr_prompt_password_finish (prompt, pair.result1, &error);
+	g_assert_error (error, G_IO_ERROR, G_IO_ERROR_CANCELLED);
+	g_clear_error (&error);
+	g_assert (password == NULL);
+
+	g_object_unref (prompt);
+	g_object_unref (pair.result1);
+	g_object_unref (pair.result2);
+	g_object_unref (connection);
+}
+
 int
 main (int argc, char **argv)
 {
+#if !GLIB_CHECK_VERSION(2,35,0)
 	g_type_init ();
+#endif
 	g_test_init (&argc, &argv, NULL);
 	g_set_prgname ("test-system-prompt");
 
@@ -566,6 +748,9 @@ main (int argc, char **argv)
 	g_test_add ("/gcr/system-prompt/properties-reset", Test, NULL, setup, test_prompt_properties_reset, teardown);
 	g_test_add ("/gcr/system-prompt/close", Test, NULL, setup, test_prompt_close, teardown);
 	g_test_add ("/gcr/system-prompt/close-cancels", Test, NULL, setup, test_close_cancels, teardown);
+	g_test_add ("/gcr/system-prompt/after-close-dismisses", Test, NULL, setup, test_after_close_dismisses, teardown);
+	g_test_add ("/gcr/system-prompt/close-from-prompter", Test, NULL, setup, test_close_from_prompter, teardown);
+	g_test_add ("/gcr/system-prompt/watch-cancels", Test, NULL, setup, test_watch_cancels, teardown);
 
 	return egg_tests_run_with_loop ();
 }
